@@ -1,6 +1,6 @@
 /* shell.c -- GNU's idea of the POSIX shell specification. */
 
-/* Copyright (C) 1987-2017 Free Software Foundation, Inc.
+/* Copyright (C) 1987-2024 Free Software Foundation, Inc.
 
    This file is part of GNU Bash, the Bourne Again SHell.
 
@@ -62,8 +62,8 @@
 #include "jobs.h"
 #else
 extern int running_in_background;
-extern int initialize_job_control __P((int));
-extern int get_tty_state __P((void));
+extern int initialize_job_control (int);
+extern int get_tty_state (void);
 #endif /* JOB_CONTROL */
 
 #include "input.h"
@@ -72,6 +72,8 @@ extern int get_tty_state __P((void));
 
 #if defined (USING_BASH_MALLOC) && defined (DEBUG) && !defined (DISABLE_MALLOC_WRAPPERS)
 #  include <malloc/shmalloc.h>
+#elif defined (MALLOC_DEBUG) && defined (USING_BASH_MALLOC)
+#  include <malloc/shmalloc.h>
 #endif
 
 #if defined (HISTORY)
@@ -92,7 +94,7 @@ extern int get_tty_state __P((void));
 #endif
 
 #if !defined (HAVE_GETPW_DECLS)
-extern struct passwd *getpwuid ();
+extern struct passwd *getpwuid (uid_t);
 #endif /* !HAVE_GETPW_DECLS */
 
 #if !defined (errno)
@@ -115,7 +117,8 @@ COMMAND *global_command = (COMMAND *)NULL;
 /* Information about the current user. */
 struct user_info current_user =
 {
-  (uid_t)-1, (uid_t)-1, (gid_t)-1, (gid_t)-1,
+  (uid_t)-1, (uid_t)-1, (uid_t)-1,
+  (gid_t)-1, (gid_t)-1, (gid_t)-1,
   (char *)NULL, (char *)NULL, (char *)NULL
 };
 
@@ -131,6 +134,9 @@ char *current_host_name = (char *)NULL;
  */
 int login_shell = 0;
 
+/* Non-zero if this shell is being run by `su'. */
+int su_shell = 0;
+
 /* Non-zero means that at this moment, the shell is interactive.  In
    general, this means that the shell is at this moment reading input
    from the keyboard. */
@@ -159,12 +165,17 @@ int autocd = 0;
 int startup_state = 0;
 int reading_shell_script = 0;
 
+int ssh_reading_startup_files = 0;
+
 /* Special debugging helper. */
 int debugging_login_shell = 0;
 
 /* The environment that the shell passes to other commands. */
 char **shell_environment;
 
+/* Non-zero when we are parsing a command, managed by parse_command/parse_comsub */
+int parsing_command = 0;
+
 /* Non-zero when we are executing a top-level command. */
 int executing = 0;
 
@@ -179,6 +190,7 @@ char *shell_name = (char *)NULL;
 
 /* time in seconds when the shell was started */
 time_t shell_start_time;
+struct timeval shellstart;
 
 /* Are we running in an emacs shell window? */
 int running_under_emacs;
@@ -190,15 +202,12 @@ int have_devfd = HAVE_DEV_FD;
 int have_devfd = 0;
 #endif
 
-/* The name of the .(shell)rc file. */
-static char *bashrc_file = DEFAULT_BASHRC;
+/* The name of the .(shell)rc file, DEFAULT_BASHRC is sourced by default */
+static char *bashrc_file;
 
 /* Non-zero means to act more like the Bourne shell on startup. */
 static int act_like_sh;
 
-/* Non-zero if this shell is being run by `su'. */
-static int su_shell;
-
 /* Non-zero if we have already expanded and sourced $ENV. */
 static int sourced_env;
 
@@ -219,8 +228,10 @@ int no_line_editing = 0;	/* non-zero -> don't do fancy line editing. */
 #else
 int no_line_editing = 1;	/* can't have line editing without readline */
 #endif
+#if defined (TRANSLATABLE_STRINGS)
 int dump_translatable_strings;	/* Dump strings in $"...", don't execute. */
 int dump_po_strings;		/* Dump strings in $"..." in po format */
+#endif
 int wordexp_only = 0;		/* Do word expansion only */
 int protected_mode = 0;		/* No command substitution with --wordexp */
 
@@ -245,8 +256,10 @@ static const struct {
 #if defined (DEBUGGER)
   { "debugger", Int, &debugging_mode, (char **)0x0 },
 #endif
+#if defined (TRANSLATABLE_STRINGS)
   { "dump-po-strings", Int, &dump_po_strings, (char **)0x0 },
   { "dump-strings", Int, &dump_translatable_strings, (char **)0x0 },
+#endif
   { "help", Int, &want_initial_help, (char **)0x0 },
   { "init-file", Charp, (int *)0x0, &bashrc_file },
   { "login", Int, &make_login_shell, (char **)0x0 },
@@ -281,10 +294,8 @@ char **subshell_envp;
 
 char *exec_argv0;
 
-#if defined (BUFFERED_INPUT)
 /* The file descriptor from which the shell is reading input. */
 int default_buffered_input = -1;
-#endif
 
 /* The following two variables are not static so they can show up in $-. */
 int read_from_stdin;		/* -s flag supplied */
@@ -303,39 +314,44 @@ static FILE *default_input;
 static STRING_INT_ALIST *shopt_alist;
 static int shopt_ind = 0, shopt_len = 0;
 
-static int parse_long_options __P((char **, int, int));
-static int parse_shell_options __P((char **, int, int));
-static int bind_args __P((char **, int, int, int));
+static int parse_long_options (char **, int, int);
+static int parse_shell_options (char **, int, int);
+static int bind_args (char **, int, int, int);
 
-static void start_debugger __P((void));
+static void start_debugger (void);
 
-static void add_shopt_to_alist __P((char *, int));
-static void run_shopt_alist __P((void));
+static void add_shopt_to_alist (char *, int);
+static void run_shopt_alist (void);
 
-static void execute_env_file __P((char *));
-static void run_startup_files __P((void));
-static int open_shell_script __P((char *));
-static void set_bash_input __P((void));
-static int run_one_command __P((char *));
+static void execute_env_file (char *);
+static void execute_profile_file (void);
+static void execute_bashrc_file (void);
+static void run_startup_files (void);
+static int open_shell_script (char *);
+static void set_bash_input (void);
+static int run_one_command (char *);
 #if defined (WORDEXP_OPTION)
-static int run_wordexp __P((char *));
+static int run_wordexp (char *);
 #endif
 
-static int uidget __P((void));
+static int uidget (void);
 
-static void init_interactive __P((void));
-static void init_noninteractive __P((void));
-static void init_interactive_script __P((void));
+static void set_option_defaults (void);
+static void reset_option_defaults (void);
 
-static void set_shell_name __P((char *));
-static void shell_initialize __P((void));
-static void shell_reinitialize __P((void));
+static void init_interactive (void);
+static void init_noninteractive (void);
+static void init_interactive_script (void);
 
-static void show_shell_usage __P((FILE *, int));
+static void set_shell_name (char *);
+static void shell_initialize (void);
+static void shell_reinitialize (void);
+
+static void show_shell_usage (FILE *, int);
 
 #ifdef __CYGWIN__
 static void
-_cygwin32_check_tmp ()
+_cygwin32_check_tmp (void)
 {
   struct stat sb;
 
@@ -352,28 +368,24 @@ _cygwin32_check_tmp ()
 #if defined (NO_MAIN_ENV_ARG)
 /* systems without third argument to main() */
 int
-main (argc, argv)
-     int argc;
-     char **argv;
+main (int argc, char **argv)
 #else /* !NO_MAIN_ENV_ARG */
 int
-main (argc, argv, env)
-     int argc;
-     char **argv, **env;
+main (int argc, char **argv, char **env)
 #endif /* !NO_MAIN_ENV_ARG */
 {
   register int i;
-  int code, old_errexit_flag, old_onecmd;
+  int code, old_errexit_flag;
 #if defined (RESTRICTED_SHELL)
   int saverst;
 #endif
   volatile int locally_skip_execution;
   volatile int arg_index, top_level_arg_index;
-#ifdef __OPENNT
+#if defined (__OPENNT) || defined (__MVS__)
   char **env;
 
   env = environ;
-#endif /* __OPENNT */
+#endif /* __OPENNT || __MVS__ */
 
   USE_VAR(argc);
   USE_VAR(argv);
@@ -429,16 +441,14 @@ main (argc, argv, env)
   arg_index = 1;
   if (arg_index > argc)
     arg_index = argc;
-  command_execution_string = shell_script_filename = (char *)NULL;
+  command_execution_string = shell_script_filename = NULL;
   want_pending_command = locally_skip_execution = read_from_stdin = 0;
   default_input = stdin;
-#if defined (BUFFERED_INPUT)
   default_buffered_input = -1;
-#endif
 
   /* Fix for the `infinite process creation' bug when running shell scripts
      from startup files on System V. */
-  login_shell = make_login_shell = 0;
+  login_shell = make_login_shell = su_shell = 0;
 
   /* If this shell has already been run, then reinitialize it to a
      vanilla state. */
@@ -455,7 +465,9 @@ main (argc, argv, env)
 
   shell_environment = env;
   set_shell_name (argv[0]);
-  shell_start_time = NOW;	/* NOW now defined in general.h */
+
+  gettimeofday (&shellstart, 0);
+  shell_start_time = shellstart.tv_sec;
 
   /* Parse argument flags from the input line. */
 
@@ -490,11 +502,13 @@ main (argc, argv, env)
 
   set_login_shell ("login_shell", login_shell != 0);
 
+#if defined (TRANSLATABLE_STRINGS)
   if (dump_po_strings)
     dump_translatable_strings = 1;
 
   if (dump_translatable_strings)
     read_but_dont_execute = 1;
+#endif
 
   if (running_setuid && privileged_mode == 0)
     disable_priv_mode ();
@@ -512,7 +526,7 @@ main (argc, argv, env)
 	}
       arg_index++;
     }
-  this_command_name = (char *)NULL;
+  this_command_name = NULL;
 
   /* First, let the outside world know about our interactive status.
      A shell is interactive if the `-i' flag was given, or if all of
@@ -613,6 +627,8 @@ main (argc, argv, env)
 	gnu_error_format = 1;
     }
 
+  compat_init ();
+
   top_level_arg_index = arg_index;
   old_errexit_flag = exit_immediately_on_error;
 
@@ -622,7 +638,7 @@ main (argc, argv, env)
   code = setjmp_sigs (top_level);
   if (code)
     {
-      if (code == EXITPROG || code == ERREXIT)
+      if (code == EXITPROG || code == ERREXIT || code == EXITBLTIN)
 	exit_shell (last_command_exit_value);
       else
 	{
@@ -655,6 +671,8 @@ main (argc, argv, env)
     {
       change_flag ('i', FLAG_ON);
       interactive = 1;
+      if (forced_interactive == 0)
+	read_but_dont_execute = 0;
     }
 
 #if defined (RESTRICTED_SHELL)
@@ -686,12 +704,27 @@ main (argc, argv, env)
     arg_index = bind_args (argv, arg_index, argc, 1);	/* $1 ... $n */
 
   /* The startup files are run with `set -e' temporarily disabled. */
+  ssh_reading_startup_files = 0;		/* paranoia */
   if (locally_skip_execution == 0 && running_setuid == 0)
     {
+      char *t;
+
       old_errexit_flag = exit_immediately_on_error;
       exit_immediately_on_error = 0;
 
+      /* Temporarily set $0 while running startup files, then restore it so
+	 we get better error messages when trying to open script files. */
+      if (shell_script_filename)
+	{
+	  t = dollar_vars[0];
+	  dollar_vars[0] = exec_argv0 ? savestring (exec_argv0) : savestring (shell_script_filename);
+	}
       run_startup_files ();
+      if (shell_script_filename)
+	{
+	  free (dollar_vars[0]);
+	  dollar_vars[0] = t;
+	}
       exit_immediately_on_error += old_errexit_flag;
     }
 
@@ -731,7 +764,7 @@ main (argc, argv, env)
 	start_debugger ();
 
 #if defined (ONESHOT)
-      executing = 1;
+      executing = shell_initialized = 1;
       run_one_command (command_execution_string);
       exit_shell (last_command_exit_value);
 #else /* ONESHOT */
@@ -748,11 +781,7 @@ main (argc, argv, env)
     {
       /* In this mode, bash is reading a script from stdin, which is a
 	 pipe or redirected file. */
-#if defined (BUFFERED_INPUT)
       default_buffered_input = fileno (stdin);	/* == 0 */
-#else
-      setbuf (default_input, (char *)NULL);
-#endif /* !BUFFERED_INPUT */
       read_from_stdin = 1;
     }
   else if (top_level_arg_index == argc)		/* arg index before startup files */
@@ -807,9 +836,7 @@ main (argc, argv, env)
 }
 
 static int
-parse_long_options (argv, arg_start, arg_end)
-     char **argv;
-     int arg_start, arg_end;
+parse_long_options (char **argv, int arg_start, int arg_end)
 {
   int arg_index, longarg, i;
   char *arg_string;
@@ -862,9 +889,7 @@ parse_long_options (argv, arg_start, arg_end)
 }
 
 static int
-parse_shell_options (argv, arg_start, arg_end)
-     char **argv;
-     int arg_start, arg_end;
+parse_shell_options (char **argv, int arg_start, int arg_end)
 {
   int arg_index;
   int arg_character, on_or_off, next_arg, i;
@@ -907,7 +932,9 @@ parse_shell_options (argv, arg_start, arg_end)
 	      o_option = argv[next_arg];
 	      if (o_option == 0)
 		{
+		  set_option_defaults ();
 		  list_minus_o_opts (-1, (on_or_off == '-') ? 0 : 1);
+		  reset_option_defaults ();
 		  break;
 		}
 	      if (set_minus_o_option (on_or_off, o_option) != EXECUTION_SUCCESS)
@@ -931,7 +958,9 @@ parse_shell_options (argv, arg_start, arg_end)
 	      break;
 
 	    case 'D':
+#if defined (TRANSLATABLE_STRINGS)
 	      dump_translatable_strings = 1;
+#endif
 	      break;
 
 	    default:
@@ -953,16 +982,18 @@ parse_shell_options (argv, arg_start, arg_end)
 
 /* Exit the shell with status S. */
 void
-exit_shell (s)
-     int s;
+exit_shell (int s)
 {
   fflush (stdout);		/* XXX */
   fflush (stderr);
 
   /* Clean up the terminal if we are in a state where it's been modified. */
 #if defined (READLINE)
-  if (RL_ISSTATE (RL_STATE_TERMPREPPED) && rl_deprep_term_function)
+  if (bash_readline_initialized && RL_ISSTATE (RL_STATE_TERMPREPPED) && rl_deprep_term_function)
+{
+itrace("exit_shell: calling rl_deprep_term_function");
     (*rl_deprep_term_function) ();
+}
 #endif
   if (read_tty_modified ())
     read_tty_cleanup ();
@@ -973,7 +1004,7 @@ exit_shell (s)
     s = run_exit_trap ();
 
 #if defined (PROCESS_SUBSTITUTION)
-  unlink_fifo_list ();
+  unlink_all_fifos ();
 #endif /* PROCESS_SUBSTITUTION */
 
 #if defined (HISTORY)
@@ -1008,12 +1039,11 @@ exit_shell (s)
 /* A wrapper for exit that (optionally) can do other things, like malloc
    statistics tracing. */
 void
-sh_exit (s)
-     int s;
+sh_exit (int s)
 {
 #if defined (MALLOC_DEBUG) && defined (USING_BASH_MALLOC)
   if (malloc_trace_at_exit && (subshell_environment & (SUBSHELL_COMSUB|SUBSHELL_PROCSUB)) == 0)
-    trace_malloc_stats (get_name_for_error (), (char *)NULL);
+    trace_malloc_stats (get_name_for_error (), NULL);
   /* mlocation_write_table (); */
 #endif
 
@@ -1024,20 +1054,26 @@ sh_exit (s)
    do any more cleanup, since a subshell is created as an exact copy of its
    parent. */
 void
-subshell_exit (s)
-     int s;
+subshell_exit (int s)
 {
   fflush (stdout);
   fflush (stderr);
 
   /* Do trap[0] if defined.  Allow it to override the exit status
      passed to us. */
+  last_command_exit_value = s;
   if (signal_is_trapped (0))
     s = run_exit_trap ();
 
   sh_exit (s);
 }
 
+void
+set_exit_status (int s)
+{
+  set_pipestatus_from_exit (last_command_exit_value = s);
+}
+
 /* Source the bash startup files.  If POSIXLY_CORRECT is non-zero, we obey
    the Posix.2 startup file rules:  $ENV is expanded, and if the file it
    names exists, that file is sourced.  The Posix.2 rules are in effect
@@ -1066,8 +1102,8 @@ subshell_exit (s)
 */
 
 static void
-execute_env_file (env_file)
-      char *env_file;
+
+execute_env_file (char *env_file)
 {
   char *fn;
 
@@ -1080,17 +1116,48 @@ execute_env_file (env_file)
     }
 }
 
+/* Execute /etc/profile and one of the personal login shell initialization files. */
 static void
-run_startup_files ()
+execute_profile_file (void)
+{
+  maybe_execute_file (SYS_PROFILE, 1);
+
+  if (act_like_sh)	/* sh */
+    maybe_execute_file ("~/.profile", 1);
+  else if ((maybe_execute_file ("~/.bash_profile", 1) == 0) &&
+	   (maybe_execute_file ("~/.bash_login", 1) == 0))	/* bash */
+    maybe_execute_file ("~/.profile", 1);
+}
+
+static void
+execute_bashrc_file (void)
+{
+#ifdef SYS_BASHRC
+#  if defined (__OPENNT)
+  maybe_execute_file (_prefixInstallPath(SYS_BASHRC, NULL, 0), 1);
+#  else
+  maybe_execute_file (SYS_BASHRC, 1);
+#  endif
+#endif
+
+  if (bashrc_file)
+    maybe_execute_file (bashrc_file, 1);
+  else
+    maybe_execute_file (DEFAULT_BASHRC, 1);
+}
+
+static void
+run_startup_files (void)
 {
 #if defined (JOB_CONTROL)
   int old_job_control;
 #endif
   int sourced_login, run_by_ssh;
 
+  /* TAG:bash-5.3 andrew.gregory.8@gmail.com 2/21/2022 */
   /* get the rshd/sshd case out of the way first. */
   if (interactive_shell == 0 && no_rc == 0 && login_shell == 0 &&
-      act_like_sh == 0 && command_execution_string)
+      act_like_sh == 0 && command_execution_string && shell_level < 2)
     {
 #ifdef SSH_SOURCE_BASHRC
       run_by_ssh = (find_variable ("SSH_CLIENT") != (SHELL_VAR *)0) ||
@@ -1099,18 +1166,18 @@ run_startup_files ()
       run_by_ssh = 0;
 #endif
 
+      ssh_reading_startup_files = 0;
       /* If we were run by sshd or we think we were run by rshd, execute
 	 ~/.bashrc if we are a top-level shell. */
+#if 1	/* TAG:bash-5.3 */
       if ((run_by_ssh || isnetconn (fileno (stdin))) && shell_level < 2)
-	{
-#ifdef SYS_BASHRC
-#  if defined (__OPENNT)
-	  maybe_execute_file (_prefixInstallPath(SYS_BASHRC, NULL, 0), 1);
-#  else
-	  maybe_execute_file (SYS_BASHRC, 1);
-#  endif
+#else
+      if (isnetconn (fileno (stdin) && shell_level < 2)
 #endif
-	  maybe_execute_file (bashrc_file, 1);
+	{
+	  ssh_reading_startup_files = 1;
+	  execute_bashrc_file ();
+	  ssh_reading_startup_files = 0;
 	  return;
 	}
     }
@@ -1138,15 +1205,7 @@ run_startup_files ()
       /* Execute /etc/profile and one of the personal login shell
 	 initialization files. */
       if (no_profile == 0)
-	{
-	  maybe_execute_file (SYS_PROFILE, 1);
-
-	  if (act_like_sh)	/* sh */
-	    maybe_execute_file ("~/.profile", 1);
-	  else if ((maybe_execute_file ("~/.bash_profile", 1) == 0) &&
-		   (maybe_execute_file ("~/.bash_login", 1) == 0))	/* bash */
-	    maybe_execute_file ("~/.profile", 1);
-	}
+	execute_profile_file ();
 
       sourced_login = 1;
     }
@@ -1171,32 +1230,13 @@ run_startup_files ()
 	  /* We don't execute .bashrc for login shells. */
 	  no_rc++;
 
-	  /* Execute /etc/profile and one of the personal login shell
-	     initialization files. */
 	  if (no_profile == 0)
-	    {
-	      maybe_execute_file (SYS_PROFILE, 1);
-
-	      if (act_like_sh)	/* sh */
-		maybe_execute_file ("~/.profile", 1);
-	      else if ((maybe_execute_file ("~/.bash_profile", 1) == 0) &&
-		       (maybe_execute_file ("~/.bash_login", 1) == 0))	/* bash */
-		maybe_execute_file ("~/.profile", 1);
-	    }
+	    execute_profile_file ();
 	}
 
       /* bash */
       if (act_like_sh == 0 && no_rc == 0)
-	{
-#ifdef SYS_BASHRC
-#  if defined (__OPENNT)
-	  maybe_execute_file (_prefixInstallPath(SYS_BASHRC, NULL, 0), 1);
-#  else
-	  maybe_execute_file (SYS_BASHRC, 1);
-#  endif
-#endif
-	  maybe_execute_file (bashrc_file, 1);
-	}
+	execute_bashrc_file ();
       /* sh */
       else if (act_like_sh && privileged_mode == 0 && sourced_env++ == 0)
 	execute_env_file (get_string_value ("ENV"));
@@ -1218,8 +1258,7 @@ run_startup_files ()
    value of `restricted'.  Don't actually do anything, just return a
    boolean value. */
 int
-shell_is_restricted (name)
-     char *name;
+shell_is_restricted (char *name)
 {
   char *temp;
 
@@ -1239,8 +1278,7 @@ shell_is_restricted (name)
    Do this also if `restricted' is already set to 1; maybe the shell was
    started with -r. */
 int
-maybe_make_restricted (name)
-     char *name;
+maybe_make_restricted (char *name)
 {
   char *temp;
 
@@ -1257,6 +1295,7 @@ maybe_make_restricted (name)
       set_var_read_only ("SHELL");
       set_var_read_only ("ENV");
       set_var_read_only ("BASH_ENV");
+      set_var_read_only ("HISTFILE");
       restricted = 1;
     }
   return (restricted);
@@ -1266,22 +1305,33 @@ maybe_make_restricted (name)
 /* Fetch the current set of uids and gids and return 1 if we're running
    setuid or setgid. */
 static int
-uidget ()
+uidget (void)
 {
   uid_t u;
 
-  u = getuid ();
+  u = current_user.uid;
+
+#if HAVE_SETRESUID
+  (void) getresuid (&current_user.uid, &current_user.euid, &current_user.saveuid);
+#else
+  current_user.uid = getuid ();
+  current_user.euid = geteuid ();
+#endif
+
+#if HAVE_SETRESGID
+  (void) getresgid (&current_user.gid, &current_user.egid, &current_user.savegid);
+#else
+  current_user.gid = getgid ();
+  current_user.egid = getegid ();
+#endif
+
   if (current_user.uid != u)
     {
       FREE (current_user.user_name);
       FREE (current_user.shell);
       FREE (current_user.home_dir);
-      current_user.user_name = current_user.shell = current_user.home_dir = (char *)NULL;
+      current_user.user_name = current_user.shell = current_user.home_dir = NULL;
     }
-  current_user.uid = u;
-  current_user.gid = getgid ();
-  current_user.euid = geteuid ();
-  current_user.egid = getegid ();
 
   /* See whether or not we are running setuid or setgid. */
   return (current_user.uid != current_user.euid) ||
@@ -1289,11 +1339,19 @@ uidget ()
 }
 
 void
-disable_priv_mode ()
+disable_priv_mode (void)
 {
-  int e;
+  int e, r;
 
-  if (setuid (current_user.uid) < 0)
+  r = 0;
+#if HAVE_SETRESUID
+  if (current_user.euid != current_user.uid || current_user.saveuid != current_user.uid)
+    r = setresuid (current_user.uid, current_user.uid, current_user.uid) ;
+#else
+  if (current_user.euid != current_user.uid)
+    r = setuid (current_user.uid);
+#endif
+  if (r < 0)
     {
       e = errno;
       sys_error (_("cannot set uid to %d: effective uid %d"), current_user.uid, current_user.euid);
@@ -1302,17 +1360,28 @@ disable_priv_mode ()
 	exit (e);
 #endif
     }
-  if (setgid (current_user.gid) < 0)
+
+  r = 0;
+#if HAVE_SETRESGID
+  if (current_user.egid != current_user.gid || current_user.savegid != current_user.gid)
+    r = setresgid (current_user.gid, current_user.gid, current_user.gid);
+#else
+  if (current_user.egid != current_user.gid)
+    r = setgid (current_user.gid);
+#endif
+  if (r < 0)
     sys_error (_("cannot set gid to %d: effective gid %d"), current_user.gid, current_user.egid);
 
   current_user.euid = current_user.uid;
   current_user.egid = current_user.gid;
+
+  current_user.saveuid = current_user.uid;
+  current_user.savegid = current_user.gid;
 }
 
 #if defined (WORDEXP_OPTION)
 static int
-run_wordexp (words)
-     char *words;
+run_wordexp (char *words)
 {
   int code, nw, nb;
   WORD_LIST *wl, *tl, *result;
@@ -1328,6 +1397,7 @@ run_wordexp (words)
 	  return last_command_exit_value = 127;
 	case ERREXIT:
 	case EXITPROG:
+	case EXITBLTIN:
 	  return last_command_exit_value;
 	case DISCARD:
 	  return last_command_exit_value = 1;
@@ -1387,8 +1457,7 @@ run_wordexp (words)
 /* Run one command, given as the argument to the -c option.  Tell
    parse_and_execute not to fork for a simple command. */
 static int
-run_one_command (command)
-     char *command;
+run_one_command (char *command)
 {
   int code;
 
@@ -1406,6 +1475,7 @@ run_one_command (command)
 	  return last_command_exit_value = 127;
 	case ERREXIT:
 	case EXITPROG:
+	case EXITBLTIN:
 	  return last_command_exit_value;
 	case DISCARD:
 	  return last_command_exit_value = 1;
@@ -1413,14 +1483,12 @@ run_one_command (command)
 	  command_error ("run_one_command", CMDERR_BADJUMP, code, 0);
 	}
     }
-   return (parse_and_execute (savestring (command), "-c", SEVAL_NOHIST));
+   return (parse_and_execute (savestring (command), "-c", SEVAL_NOHIST|SEVAL_RESETLINE));
 }
 #endif /* ONESHOT */
 
 static int
-bind_args (argv, arg_start, arg_end, start_index)
-     char **argv;
-     int arg_start, arg_end, start_index;
+bind_args (char **argv, int arg_start, int arg_end, int start_index)
 {
   register int i;
   WORD_LIST *args, *tl;
@@ -1472,14 +1540,14 @@ bind_args (argv, arg_start, arg_end, start_index)
 }
 
 void
-unbind_args ()
+unbind_args (void)
 {
   remember_args ((WORD_LIST *)NULL, 1);
   pop_args ();				/* Reset BASH_ARGV and BASH_ARGC */
 }
 
 static void
-start_debugger ()
+start_debugger (void)
 {
 #if defined (DEBUGGER) && defined (DEBUGGER_START_FILE)
   int old_errexit;
@@ -1504,8 +1572,7 @@ start_debugger ()
 }
 
 static int
-open_shell_script (script_name)
-     char *script_name;
+open_shell_script (char *script_name)
 {
   int fd, e, fd_is_tty;
   char *filename, *path_filename, *t;
@@ -1551,7 +1618,7 @@ open_shell_script (script_name)
   if (exec_argv0)
     {
       free (exec_argv0);
-      exec_argv0 = (char *)NULL;
+      exec_argv0 = NULL;
     }
 
   if (file_isdir (filename))
@@ -1573,7 +1640,7 @@ open_shell_script (script_name)
   GET_ARRAY_FROM_VAR ("BASH_SOURCE", bash_source_v, bash_source_a);
   GET_ARRAY_FROM_VAR ("BASH_LINENO", bash_lineno_v, bash_lineno_a);
 
-  array_push (bash_source_a, filename);
+  push_source (bash_source_a, filename);
   if (bash_lineno_a)
     {
       t = itos (executing_line_number ());
@@ -1620,7 +1687,7 @@ open_shell_script (script_name)
 	}
       else if (sample_len > 0 && (check_binary_file (sample, sample_len)))
 	{
-	  internal_error (_("%s: cannot execute binary file"), filename);
+	  internal_error ("%s: %s", filename, _("cannot execute binary file"));
 #if defined (JOB_CONTROL)
 	  end_job_control ();	/* just in case we were run as bash -i script */
 #endif
@@ -1635,22 +1702,8 @@ open_shell_script (script_name)
      not match with ours. */
   fd = move_to_high_fd (fd, 1, -1);
 
-#if defined (BUFFERED_INPUT)
   default_buffered_input = fd;
   SET_CLOSE_ON_EXEC (default_buffered_input);
-#else /* !BUFFERED_INPUT */
-  default_input = fdopen (fd, "r");
-
-  if (default_input == 0)
-    {
-      file_error (filename);
-      exit (EX_NOTFOUND);
-    }
-
-  SET_CLOSE_ON_EXEC (fd);
-  if (fileno (default_input) != fd)
-    SET_CLOSE_ON_EXEC (fileno (default_input));
-#endif /* !BUFFERED_INPUT */
 
   /* Just about the only way for this code to be executed is if something
      like `bash -i /dev/stdin' is executed. */
@@ -1659,12 +1712,7 @@ open_shell_script (script_name)
       dup2 (fd, 0);
       close (fd);
       fd = 0;
-#if defined (BUFFERED_INPUT)
       default_buffered_input = 0;
-#else
-      fclose (default_input);
-      default_input = stdin;
-#endif
     }
   else if (forced_interactive && fd_is_tty == 0)
     /* But if a script is called with something like `bash -i scriptname',
@@ -1680,24 +1728,20 @@ open_shell_script (script_name)
 
 /* Initialize the input routines for the parser. */
 static void
-set_bash_input ()
+set_bash_input (void)
 {
   /* Make sure the fd from which we are reading input is not in
      no-delay mode. */
-#if defined (BUFFERED_INPUT)
   if (interactive == 0)
     sh_unset_nodelay_mode (default_buffered_input);
   else
-#endif /* !BUFFERED_INPUT */
     sh_unset_nodelay_mode (fileno (stdin));
 
   /* with_input_from_stdin really means `with_input_from_readline' */
   if (interactive && no_line_editing == 0)
     with_input_from_stdin ();
-#if defined (BUFFERED_INPUT)
   else if (interactive == 0)
     with_input_from_buffered_stream (default_buffered_input, dollar_vars[0]);
-#endif /* BUFFERED_INPUT */
   else
     with_input_from_stream (default_input, dollar_vars[0]);
 }
@@ -1707,10 +1751,8 @@ set_bash_input ()
    is non-zero, we close default_buffered_input even if it's the standard
    input (fd 0). */
 void
-unset_bash_input (check_zero)
-     int check_zero;
+unset_bash_input (int check_zero)
 {
-#if defined (BUFFERED_INPUT)
   if ((check_zero && default_buffered_input >= 0) ||
       (check_zero == 0 && default_buffered_input > 0))
     {
@@ -1718,13 +1760,6 @@ unset_bash_input (check_zero)
       default_buffered_input = bash_input.location.buffered_fd = -1;
       bash_input.type = st_none;		/* XXX */
     }
-#else /* !BUFFERED_INPUT */
-  if (default_input)
-    {
-      fclose (default_input);
-      default_input = (FILE *)NULL;
-    }
-#endif /* !BUFFERED_INPUT */
 }
       
 
@@ -1733,8 +1768,7 @@ unset_bash_input (check_zero)
 #endif
 
 static void
-set_shell_name (argv0)
-     char *argv0;
+set_shell_name (char *argv0)
 {
   /* Here's a hack.  If the name of this shell is "sh", then don't do
      any startup files; just try to be more like /bin/sh. */
@@ -1763,25 +1797,52 @@ set_shell_name (argv0)
     shell_name = PROGRAM;
 }
 
+/* Some options are initialized to -1 so we have a way to determine whether
+   they were set on the command line. This is an issue when listing the option
+   values at invocation (`bash -o'), so we set the defaults here and reset
+   them after the call to list_minus_o_options (). */
+/* XXX - could also do this for histexp_flag, jobs_m_flag */
 static void
-init_interactive ()
+set_option_defaults (void)
 {
-  expand_aliases = interactive_shell = startup_state = 1;
-  interactive = 1;
 #if defined (HISTORY)
-  remember_on_history = enable_history_list = 1;	/* XXX */
-  histexp_flag = history_expansion;			/* XXX */
+  enable_history_list = 0;
 #endif
 }
 
 static void
-init_noninteractive ()
+reset_option_defaults (void)
 {
 #if defined (HISTORY)
+  enable_history_list = -1;
+#endif
+}
+
+static void
+init_interactive (void)
+{
+  expand_aliases = expaliases_flag = 1;
+  interactive_shell = startup_state = interactive = 1;
+#if defined (HISTORY)
+  if (enable_history_list == -1)
+    enable_history_list = 1;				/* set default  */
+  remember_on_history = enable_history_list;
+#  if defined (BANG_HISTORY)
+  histexp_flag = history_expansion;			/* XXX */
+#  endif
+#endif
+}
+
+static void
+init_noninteractive (void)
+{
+#if defined (HISTORY)
+  if (enable_history_list == -1)			/* set default */
+    enable_history_list = 0;
   bash_history_reinit (0);
 #endif /* HISTORY */
   interactive_shell = startup_state = interactive = 0;
-  expand_aliases = posixly_correct;	/* XXX - was 0 not posixly_correct */
+  expand_aliases = expaliases_flag = posixly_correct;	/* XXX - was 0 not posixly_correct */
   no_line_editing = 1;
 #if defined (JOB_CONTROL)
   /* Even if the shell is not interactive, enable job control if the -i or
@@ -1791,17 +1852,21 @@ init_noninteractive ()
 }
 
 static void
-init_interactive_script ()
+init_interactive_script (void)
 {
-  init_noninteractive ();
-  expand_aliases = interactive_shell = startup_state = 1;
 #if defined (HISTORY)
-  remember_on_history = enable_history_list = 1;	/* XXX */
+  if (enable_history_list == -1)
+    enable_history_list = 1;
+#endif
+  init_noninteractive ();
+  expand_aliases = expaliases_flag = interactive_shell = startup_state = 1;
+#if defined (HISTORY)
+  remember_on_history = enable_history_list;	/* XXX */
 #endif
 }
 
 void
-get_current_user_info ()
+get_current_user_info (void)
 {
   struct passwd *entry;
 
@@ -1837,7 +1902,7 @@ get_current_user_info ()
 /* Do whatever is necessary to initialize the shell.
    Put new initializations in here. */
 static void
-shell_initialize ()
+shell_initialize (void)
 {
   char hostname[256];
   int should_be_restricted;
@@ -1917,7 +1982,7 @@ shell_initialize ()
    had some initialization performed.  This is supposed to reset the world
    back to a pristine state, as if we had been exec'ed. */
 static void
-shell_reinitialize ()
+shell_reinitialize (void)
 {
   /* The default shell prompts. */
   primary_prompt = PPROMPT;
@@ -1933,13 +1998,25 @@ shell_reinitialize ()
   no_rc = no_profile = 1;
 
   /* Things that get 0. */
-  login_shell = make_login_shell = interactive = executing = 0;
-  debugging = do_version = line_number = last_command_exit_value = 0;
-  forced_interactive = interactive_shell = 0;
+  login_shell = make_login_shell = su_shell = executing = 0;
+  debugging = debugging_mode = 0;
+  do_version = line_number = last_command_exit_value = 0;
+  forced_interactive = interactive_shell = interactive = 0;
   subshell_environment = running_in_background = 0;
-  expand_aliases = 0;
+  expand_aliases = expaliases_flag = 0;
   bash_argv_initialized = 0;
 
+  /* 20240120 */
+  startup_state = reading_shell_script = 0;
+  /* XXX - inherit posixly_correct? */
+
+  /* The shell has never done this. Should it? */
+#if 0
+  reset_shell_flags ();
+  reset_shell_options ();
+  reset_shopt_options ();
+#endif
+
   /* XXX - should we set jobs_m_flag to 0 here? */
 
 #if defined (HISTORY)
@@ -1952,7 +2029,7 @@ shell_reinitialize ()
 
   /* Ensure that the default startup file is used.  (Except that we don't
      execute this file for reinitialized shells). */
-  bashrc_file = DEFAULT_BASHRC;
+  bashrc_file = NULL;
 
   /* Delete all variables and functions.  They will be reinitialized when
      the environment is parsed. */
@@ -1969,9 +2046,7 @@ shell_reinitialize ()
 }
 
 static void
-show_shell_usage (fp, extra)
-     FILE *fp;
-     int extra;
+show_shell_usage (FILE *fp, int extra)
 {
   int i;
   char *set_opts, *s, *t;
@@ -1989,7 +2064,11 @@ show_shell_usage (fp, extra)
 
   for (i = 0, set_opts = 0; shell_builtins[i].name; i++)
     if (STREQ (shell_builtins[i].name, "set"))
-      set_opts = savestring (shell_builtins[i].short_doc);
+      {
+	set_opts = savestring (shell_builtins[i].short_doc);
+	break;
+      }
+
   if (set_opts)
     {
       s = strchr (set_opts, '[');
@@ -2016,9 +2095,7 @@ show_shell_usage (fp, extra)
 }
 
 static void
-add_shopt_to_alist (opt, on_or_off)
-     char *opt;
-     int on_or_off;
+add_shopt_to_alist (char *opt, int on_or_off)
 {
   if (shopt_ind >= shopt_len)
     {
@@ -2031,7 +2108,7 @@ add_shopt_to_alist (opt, on_or_off)
 }
 
 static void
-run_shopt_alist ()
+run_shopt_alist (void)
 {
   register int i;
 
