@@ -1,6 +1,6 @@
 /* version.c -- distribution and version numbers. */
 
-/* Copyright (C) 1989-2018 Free Software Foundation, Inc.
+/* Copyright (C) 1989-2024 Free Software Foundation, Inc.
 
    This file is part of GNU Bash, the Bourne Again SHell.
 
@@ -18,12 +18,16 @@
    along with Bash.  If not, see <http://www.gnu.org/licenses/>.
 */
 
-#include <config.h>
+/* assume C90/POSIX-1992 compilation environment if cross-compiling */
+
+#ifndef CROSS_COMPILING
+#  include <config.h>
+#else
+#  include <buildconf.h>
+#endif
 
 #include <stdio.h>
 
-#include "stdc.h"
-
 #include "version.h"
 #include "patchlevel.h"
 #include "conftypes.h"
@@ -43,21 +47,22 @@ const char * const release_status = (char *)0;
 #endif
 const char * const sccs_version = SCCSVERSION;
 
-const char * const bash_copyright = N_("Copyright (C) 2018 Free Software Foundation, Inc.");
+const char * const bash_copyright = N_("Copyright (C) 2024 Free Software Foundation, Inc.");
 const char * const bash_license = N_("License GPLv3+: GNU GPL version 3 or later <http://gnu.org/licenses/gpl.html>\n");
 
 /* If == 31, shell compatible with bash-3.1, == 32 with bash-3.2, and so on */
+const int default_compatibility_level = DEFAULT_COMPAT_LEVEL;
 int shell_compatibility_level = DEFAULT_COMPAT_LEVEL;
 
 /* Functions for getting, setting, and displaying the shell version. */
 
 /* Forward declarations so we don't have to include externs.h */
-extern char *shell_version_string __P((void));
-extern void show_shell_version __P((int));
+extern char *shell_version_string (void);
+extern void show_shell_version (int);
 
 /* Give version information about this shell. */
 char *
-shell_version_string ()
+shell_version_string (void)
 {
   static char tt[32] = { '\0' };
 
@@ -80,8 +85,7 @@ shell_version_string ()
 }
 
 void
-show_shell_version (extended)
-     int extended;
+show_shell_version (int extended)
 {
   printf (_("GNU bash, version %s (%s)\n"), shell_version_string (), MACHTYPE);
   if (extended)
